/*
 * fcomp.c
 * Fox Compression
 * GPL license
 */
/*
Here are the specs for both formats.

fcomp-standard:
TYPE   SIZE    DESC
byte      1    Pointer to the next backpointer.  0 is EOS
stream    X    a bunch of bytes to take as pure stream
byte      1    Distance.  It is 1-256 bytes back look (not 0-255)
byte      1    Length.  Size to copy, 0-255.  0 allows for chaining
               backpointers in the stream when there is a lot of nonredundant
	       data

This repeats until EOS (which is likely a 0 0 0 if the end of the stream was
not found in the previous 256 bytes; if it was, then it's just D L 0).

fcomp-extended:
TYPE   SIZE    DESC
byte      1    0.  This indicates EOS in an fcomp-standard, thus since this
               makes absolutely no sense it signifies an fcomp-extended.
byte      1    Tells how many bytes the pointers are.  One only: beginning of
               the stream.  This is N
N         N    Tell what the window is.  This is the max distance for a
               backpointer.
N         N    Distance forward to next backpointer.  N bytes in size. 0 is EOS
stream    X    A bunch of bytes to take as pure data stream
N         N    Distance back to go to get data.  Base 1
N         N    Length of data to copy.  Base 0

The first three are the header (2+N bytes); the rest repeat until EOS.
*/

/*
 * How to use:
 * 1. Block-style:  Pass data until the output stream is about X in size, then
 *    kill it, write it, make a new one, and start all over with the remaining
 *    data.
 * 2. Full-style:  Pass data until you run out of input data.
 */
#include <stdint.h>
#include "fcomp.h"
#include <malloc.h>

// void* fcomp_resize(void*&, int32_t, int32_t); // resize
int32_t fcomp_scan(int32_t lob, char* buf, int32_t loxb, char* xbuf, int32_t start);
  // Scans xbuf for buf and returns the offset of the first occurance.

fcomp* fcomp_new()
{ //    Makes a new fcomp object and returns a pointer (no c++).
  fcomp*  fc;
  fc = (fcomp*)malloc(sizeof(fcomp));
  fc->flags = 0x00; // No flags
  fc->psize = 0x00; // not determined YET.
  fc->mdist = 0xFF; // Max distance defaults to 256 regardless.
  fc->lob   = 0x00; // no buffer yet.
  fc->stream= NULL; // no stream yet either.  This will be lob in size
  fc->abuf  = NULL; // No analysis buffer.
  fc->last_pointer = 0x00;
  fc->rdist = 0x00;
  return fc;
}

fdecomp* fdecomp_new()
{ //    Makes a new fdecomp object and returns a pointer (no c++).
  fdecomp* fd;
  fd = (fdecomp*)malloc(sizeof(fdecomp));
  fd->flags = 0x00;
  fd->psize = 0x00;
  fd->mdist = 0xFF; // Assume this.
  fd->rdist = 0x00; // real size
  fd->lob   = 0x00;
  fd->lop   = 0x00;
  fd->stream= NULL;
  fd->abuf  = NULL; // analysis
  fd->pbuf  = NULL; // processing
  fd->dnext = 0x00;
  return fd;
}

#warning [FIXME] fcomp_push():  Need to finish it.  Read the comments
int32_t fcomp_push(fcomp* hInst, int32_t lob, char* stream)
{ //    stream to compress and put onto end of fcomp.
  int32_t i;
  int32_t j;
  int32_t k; // stream position
  if (hInst->flags & FC_LOCK) // Locked?  ... DO NOT THREAD!
    return FCF_ERROR; // Error
  if (lob == 0)
    return 0; // quietly ignore the no-data requests.
  hInst->flags |= FC_LOCK; // lock it.
  
  if (hInst->psize = 0)
  { // pointer size is zero.  Assume standard.
    hInst->psize = 1;
    hInst->mdist = 0xFF;
  }
  
  if (hInst->abuf == NULL) // New abuf
  {
    hInst->abuf = (char*)malloc(hInst->mdist + 1);
    hInst->rdist = 0;
  }
  
  // Here, what we need to do is:
  // 1. Check to see if the next byte of the buffer is in the abuf
  // 2. If it is, note so, as how far to point back and how far to copy. If
  //    not, then just put the byte on the output stream.
  // 3. Put the next byte on the abuf
  // 4. Keep counting up copy distance until mdist is reached or there is
  //    a difference.  If you find a string that repeats more than
  //    (psize * 3) bytes, put pointers on the output stream and change the
  //    previous pointer in the output stream to reflect the location of
  //    this one.  If you go for mdist bytes without another pointer, make
  //    one of 0 0 X, and point the previous pointer at it.
  //    If the string is <= (psize * 3) bytes, just put the data on the output
  //    stream.
  //
  //    This may require some extra work, i.e. find multiple matches, shove 'em
  //    in a linked list, start removing linked list elements as matches end.
  //    This way you could quickly find the best compression.  This WOULD
  //    require a little extra in the fcomp struct, and a new linked list
  //    struct.
  
  k = 0; // Reset stream position
  for (i = 0; i < lob; i++)
  {
    // Step 1 goes here
    // Step 2 goes here
    // Step 3:
    if (hInst->rdist == hInst->mdist)
    {
      for (j = 0; j < hInst->mdist; j++)
      { // move it back
        hInst->abuf[j] = hInst->abuf[j+1];
      }
    }
    hInst->abuf[hInst->rdist] = stream[k];
    k++;
    // step 4 must be integrated throughout this loop.
  }
  
  
  hInst->flags &= ~FC_LOCK; // Unlock it.
  return 0;
}

char* fcomp_pull(fcomp* hInst, int32_t lob)
{ //    Pulls compressed data from the beginning of the fcomp and returns it.  If
//   the buffer is not lob long, it returns 0.
  char* outbuf;
  int32_t i;
  int32_t j;
  if (hInst->flags & FC_LOCK) // Locked?  ... DO NOT THREAD!
    return (char*)FCF_ERROR; // Error
  if (lob == 0)
    return 0; // FEH!  ASK FOR SOMETHING!
  hInst->flags |= FC_LOCK; // Lock it.
  
  outbuf = (char*)malloc(lob);
  for (i = 0; i < lob; i++)
  {
    outbuf[i] = hInst->stream[i];
  }
  j = i;
  for (; i < hInst->lob; i++)
  { // Move it back
    hInst->stream[i - j] = hInst->stream[i];
  }
  realloc(hInst->stream, hInst->lob - lob); // Resize it now.
  hInst->flags &= ~FC_LOCK; // unlock
  return outbuf;
}

int32_t fcomp_size(fcomp* hInst)
{ //    Returns hInst->last_pointer.  Make sure pull doesn't break the stream.
  return hInst->last_pointer;
}

int32_t fcomp_delete(fcomp* hInst)
{ //    delete an fcomp.
  if (hInst->flags & FC_LOCK) // Locked?  ... DO NOT THREAD!
    return FCF_ERROR; // Error

  if (hInst == NULL)
    return 0; // Feh.
  if (hInst->stream != NULL)
    free(hInst->stream);
  if (hInst->abuf != NULL)
    free(hInst->abuf);
  free(hInst);
  return 0;
}

#warning [FIXME] fdecomp_push():  Only works with fcomp-standard streams (8 bit pointer assumed).  Read the comments.  fcomp-standard is recommended over fcomp-extended for speed
int32_t fdecomp_push(fdecomp* hInst, int32_t lob, char* stream)
{ //    stream to decompress and put onto end of fdecomp.
  // Logic is as follows
  // 1.  Prepend the Processing Buffer
  // 2.  Begin analysis
  // 3.  If you reach the next pointer, read its info in.  Begin to copy the
  //     data.  If you hit the end of the abuf, push up to mdist bytes of the
  //     input stream on it, then continue to copy.  Pointers do NOT overlap!
  //     (i.e. you have to copy in X distance, but you won't hit the next
  //     pointer).  Remember to stick the new dnext into the hInst
  // 4.  When not processing a pointer, just copy over however many bytes are
  //     left until the next pointer is reached, then do step 3
  // 5.  return 0
  int32_t i;
  int32_t j;
  int32_t stream_offset = 0; // add the size of all pointers.
  int32_t dback;
  int32_t lcopy;
  char    a;
  if (hInst->flags & FC_LOCK) // Locked?  ... DO NOT THREAD!
    return FCF_ERROR; // Error
  if (lob == 0 || hInst->flags & FC_EOS) // no buffer or end of stream already
    return 0; // Asked to do nothing.  Done successfully.
  hInst->flags |= FC_LOCK; // Lock.

// this is code solely for fcomp-standard.
  if (hInst->pbuf != NULL) // something on pbuf?
  { // Recreate stream from pbuf by prepending pbuf to stream
    realloc(hInst->pbuf, hInst->lop + lob); // make it bigger
    for (i = 0; i < lob; i++) // copy over
    {
      hInst->pbuf[hInst->lop + i] = stream[i];
    }
    free(stream); // ditch stream
    lob = lob + hInst->lop; // recalculate length
    stream = hInst->pbuf; // Change pointers,
    hInst->pbuf = NULL; // and we're done
  }
  i = 0;
  if (hInst->psize == 0) // Not yet started
  {  // Adjust this for psize difference:
    // if input first byte == 0 then is a fcomp-extended
    hInst->psize = 1; // 8 bit.
    hInst->abuf = (char*)malloc(hInst->mdist); // Max distance!
    hInst->dnext = stream[i++]; // load dist to first pointer!
  }
  for (; i < lob; i++) // i is set already.  don't mess with this.
  {
    hInst->dnext--; // Decriment distance
    if (hInst->dnext <= 0) // WOW! POINTER!
    {
      if (lob - i <= 2) // We're outta buffer though!
      {// slap what's left onto the 
        hInst->lop  = lob - i;
        hInst->pbuf = (char*)malloc(hInst->lop);
	for (j = 0; j < lob - i; j++)
	{ // copy the rest
	  hInst->pbuf[j] = stream[i + j + stream_offset]; 
	}
        return 0; // This is perfectly normal
      } // Else, process the pointer!
      hInst->dnext = stream[i++ + stream_offset];
      dback = stream[i++ + stream_offset]; // distance back
      lcopy = stream[i++ + stream_offset]; // and length to copy
      // copy data
      // Look that far back, realloc the stream to lcopy bigger, start
      // copying.
      // Since this puts data on the stream, and subsiquently on the
      // abuf analysis buffer, there is no chance that at this point we
      // will need to stop and continue when another piece of the stream
      // is added.
      hInst->lob += lcopy; // increase size...
      realloc(hInst->stream, hInst->lob); // and alloc more
      for (; lcopy > 0; lcopy--)
      { // just decrease lcopy to 0
        a = hInst->abuf[hInst->rdist - (dback + 1)]; // grab the byte
        if (hInst->rdist == hInst->mdist)
	{ // ahh crap gotta move this
	  for (j = 0; j < hInst->mdist - 1; j++)
	  {
	    hInst->abuf[j] = hInst->abuf[j+1];
	  }
	  hInst->abuf[hInst->mdist] = a; // Putith the new byte at the end
	}
	else
	{ // ahh, still packing abuf
	  hInst->abuf[hInst->rdist] = a; // Put it
	  hInst->abuf++; // and expand it
	  dback--; // and move with it.
	}
	realloc(hInst->stream, hInst->lob + 1); // resize
	hInst->stream[hInst->lob] = a; // copy
	hInst->lob++; // reflect size in the length
      } /*for (; lcopy > 0; lcopy++)*/
      stream_offset += 3; // offset due to pointers.  LAST THING TO DO!
      if (hInst->dnext == 0) // EOS
      {
        hInst->flags |= FC_EOS;
        return 0; // End of stream.  Leave regardless!
      }
    } /* if (hInst->dnext <= 0), if (we hit a pointer) */
    else
    { // Okay, just copy data then.
      hInst->lob++; // increase buffer size
      realloc(hInst->stream, hInst->lob);
      hInst->stream[hInst->lob - 1] = stream[i + stream_offset];
    }
  }

  hInst->flags &= ~FC_LOCK; // Unlock
  return 0;
}

char* fdecomp_pull(fdecomp* hInst, int32_t lob)
{ //    Pulls uncompressed data from the beginning of the fdecomp and returns it.
//    If the buffer is not lob long, it returns 0.
  char* outbuf;
  int32_t i;
  int32_t j;
  if (hInst->flags & FC_LOCK) // Locked?  ... DO NOT THREAD!
    return (char*)FCF_ERROR; // Error
  if (lob == 0)
      return 0; // FEH!  ASK FOR SOMETHING!
  hInst->flags |= FC_LOCK; // lock it
  
  outbuf = (char*)malloc(lob);
  for (i = 0; i < lob; i++)
  {
    outbuf[i] = hInst->stream[i];
  }
  j = i;
  for (; i < hInst->lob; i++)
  { // Move it back
    hInst->stream[i - j] = hInst->stream[i];
  }
  realloc(hInst->stream, hInst->lob - lob); // Resize it now.

  hInst->flags &= ~FC_LOCK; // unlock
  return outbuf;
}

int32_t fdecomp_size(fdecomp* hInst)
{ //     Returns hInst->lob.  There is no reason to horde any uncompressed data.
  return hInst->lob;
}

int32_t fdecomp_delete(fdecomp* hInst)
{  //  delete an fdecomp.
  if (hInst->flags & FC_LOCK) // Locked?  ... DO NOT THREAD!
    return FCF_ERROR; // Error
  
  if (hInst->stream != NULL)
    free(hInst->stream);
  if (hInst->abuf != NULL)
    free(hInst->abuf);
  if (hInst->pbuf != NULL)
    free(hInst->pbuf);
  free(hInst);
  return 0;
}

// Depricated
/*
void* fcomp_resize(void*& buf, int32_t size, int32_t newsize)
{ // NULL return means nothing done
  void* ybuf = 0;
  int32_t i = size;
  int32_t j;
  if (!(ybuf = malloc(newsize))) // can't alloc
    return NULL; // give null
  if (newsize < i)
    i = newsize; // how much to copy
  if (buf != NULL)
    for (j = 0; j < i; j++)
    { // copy what ye may.  The rest is left as garbage.
      ((char*)ybuf)[j] = ((char*)buf)[j];
    }
    free(buf); // free buf
  }
  buf = ybuf; // Transfer
  return buf; // return it
}
*/

// Now the scan functions.
int32_t BruteForce(int32_t lob, char* buf, int32_t loxb, char* xbuf, int32_t start);
int32_t bmbinary(int32_t lob, char* buf, int32_t loxb, char* xbuf, int32_t start);
int32_t kmp(int32_t lob, char* buf, int32_t loxb, char* xbuf, int32_t start);

int32_t fcomp_scan(int32_t lob, char* buf, int32_t loxb, char* xbuf, int32_t start)
{ // Select a scanning algorithm, make fcomp_scan call it.
  // I suggest using the BruteForce() algorithm if you are testing anythnig other
  // than the string search algorithms!

   return BruteForce(lob, buf, loxb, xbuf, start); // This is the most stable
// return bmbinary(lob, buf, loxb, xbuf, start); // This is the fastest
// return kmp(lob, buf, loxb, xbuf, start); // This needs to be coded

}

int32_t BruteForce(int32_t lob, char* buf, int32_t loxb, char* xbuf, int32_t start)
{ // Brute force string search.  Slow, but it's so easy to code (i.e. no bugs!)
  int32_t i = start;
  int32_t j;
  if (i + lob > loxb)
    return ~0; // BAD! BAD BAD BAD!  Searching past the end!  No match!
  for ( ; i + lob <= loxb; i++)
  {
    for(j = 0; j < lob && buf[j] == xbuf[i+j]; j++)
    { // heh.  Nothing to do.
    }
    if (j == lob) // match has been found!
      return i;
  }
  return ~0; // no match found x.x
}

int32_t bmbinary(int32_t lob, char* buf, int32_t loxb, char* xbuf, int32_t start)
{ // Ahh, boyer-moore.  It took me a year to get this right for text scans.  Use
  // if you're sure it's right.
  int32_t i=0;
  int32_t j=lob; // Pattern to look for
  int32_t l=0;
  int32_t m=0;
  int32_t n=0;
  int32_t o=loxb; // search string
  int32_t* k=0;   //bnnbb
  n=start; // bobnb
  k = (int32_t*)malloc(sizeof(int32_t) * 256); // initialize
  for (i = 0; i < 256; i++)
    k[i] = j; // Jump past
  for (i = 0; i <= j; i++)
  { // Build lookup table
    k[buf[i]] = j - i; // length - offset
    if (k[buf[i]] == 0)
    { // NO!  No moving zero.  This is a failsafe for if my logic is bad :-p
      k[buf[i]] = 1;
    }
  }
  i = 0;
  while (1)
  {
    for (m = j; m >= 0 && (n+m) < o; m--)
    { //  Scan
      if (buf[m] != xbuf[m+n])
      {
        n += k[xbuf[m+n]]; // Shift off
        break; // exit for
      }
    }
    if (m == -1)
    { // For hit end
     break;
    }
    if ((n+m) > o - 1)
    { // WHOA!!! Past end of string!!!!!
      n = -1; // No match!!#(!@$&!$
      break; // Let's not stomp RAM
    }
  }
  free(k);
  return n;
}

int32_t kmp(int32_t lob, char* buf, int32_t loxb, char* xbuf, int32_t start)
{ // Knuth-Moris-Pratt.  Well it's very different from what's on Moore's page
  // it seems.  I'm not writing this one!
}

/*
As you can see the API is quite simple.  both new()'s malloc new fcomps and
initialize them.  Both delete's free the data properly.  To use this, just push
the input data, ask for the size, and pull the output data.  Pulling is
destructive, i.e. the data is actually removed from the object.

fdecomp's automatically identify fcomp-extended streams.
*/
// This is the original code for bmsearch from fdber, strings only!
/*
extern "C" STDCALL long bmsearch(
char* spat,
char* strscan,
int startat
)
{ // C-style Boyer-moore search
  int i=0;
  int j=0;
  int l=0;
  int m=0;
  int n=0;
  int* k=0;   //bnnbb
  n=startat; // bobnb
  j = strlen(spat) - 1; // Pattern
  k = new int[256]; // initialize
  for (i = 0; i < 256; i++)
    k[i] = j; // Jump past
  for (i = 0; i <= j; i++)
  { // Build lookup table
    k[spat[i]] = (j) - i; // length - offset 
    if (k[spat[i]] == 0)
    { // NO!  No moving zero
      k[spat[i]] = 1;
    }
  }
  i = 0;
  while (1)
  {
    for (m = j; m >= 0 && (n+m) < (strlen(strscan)); m--)
    { //  Scan
      if (spat[m] != strscan[m+n])
      {
        n += k[strscan[m+n]]; // Shift off
        break; // exit for
      }
    }
    if (m == -1)
    { // For hit end
     break;
    }
    if ((n+m) > strlen(strscan) - 1)
    { // WHOA!!! Past end of string!!!!!
      n = -1; // No match!!#(!@$&!$
      break; // Let's not stomp RAM
    }
  }
  delete k;
  return n;
}
*/
